#!/usr/bin/env python

"""sqlnub.py - very light sqlite3 client for interfacing with a data file

Should be light enough for SQL novices, with just enough features to help a
programmer get off the ground with a mini database.

Suitable for high scores, settings, resource management, persistent game
state, and light data storage and crunching.
"""

__version__ = '1.0.0'
__vernum__ = __version__.split('.')

"""
References:
https://docs.python.org/2/library/sqlite3.html
http://www.tutorialspoint.com/sqlite/
http://www.w3schools.com/sql/

Tools:
http://www.gfi.com/blog/top-10-free-database-tools-for-sys-admins/

Helpful info:
http://zetcode.com/db/sqlite/datamanipulation/
https://www.sqlite.org/docs.html
http://stackoverflow.com/questions/11996394/is-there-a-way-to-get-a-schema-of-a-database-from-within-python
http://stackoverflow.com/questions/82875/how-do-i-list-the-tables-in-a-sqlite-database-file
http://www.openbookproject.net/courses/python4fun/sqlschool.html
"""

import sqlite3
import traceback


class SqlNub(object):
    """SqlNub - a super-lightweight sqlite3 interface for basic use cases

    This class is intended for SQL nubs like me. Although the entry level
    is very low, you may have to do some reading in the materials listed in
    the docstring accompanying the module.

    There is enough here to create a database with some tables, and work
    with some data.

    This class provides table creation, querying table and column info,
    executing statements, getting a cursor, rolling back and committing
    transactions, and closing the database.

    The row_factory is set to sqlite3.Row to enable the nicer features of the
    class. Columns can be accessed by name:
    'cid', 'name', 'type', 'notnull', 'dflt_value', 'pk'.

    Do not forget to commit() your transactions.

    More advanced sqlite3 features can be accessed by directly calling the
    methods on the self.conn object, the methods on the object returned by
    self.cursor(), and members of the sqlite3 module. The SqlNub class is
    easily extensible: there is only one attribute, self.conn.

    All methods may raise:
        InternalError, ProgrammingError, DataError, NotSupportedError,
        IntegrityError, Error, InterfaceError, OperationalError,
        DatabaseError

    """

    def __init__(self, filename):
        self.conn = sqlite3.connect(filename)
        self.conn.row_factory = sqlite3.Row

    def create_table(self, name, columns):
        """create a table with specified name and columns

        :param name: string; table name
        :param columns: tuple; (('column_name', 'data_type')[, ...]')
        :return: sqlite3.Cursor

        Example:
        db.create_table('high_scores', (('player', 'TEXT'), ('score', 'INT'))

        See also:
        https://www.sqlite.org/lang_createtable.html
        https://www.sqlite.org/datatype3.html#affinity
        """
        cols = ', '.join(['{0} {1}'.format(n, t) for n, t in columns])
        sql = 'CREATE TABLE {0} ({1})'.format(name, cols)
        try:
            return self.conn.execute(sql)
        except sqlite3.Error as e:
            print('attempted sql: {}'.format(sql))
            raise e


    def get_db_info(self):
        """get a dict describing tables and columns

        :return: dict; see below

        Format:
        {
            table_name: {
                column_name: tuple(column_info...),
                ...
            },
            ...
        }
        """
        tables = {}
        for table_name in self.get_table_names():
            if table_name not in tables:
                tables[table_name] = {}
            columns = tables[table_name]
            for column_info in self.get_column_info(table_name):
                columns[column_info['name']] = column_info
        return tables

    def get_table_names(self):
        """names of all user tables in a database

        :return: list; list of string

        Example return value:
        (u'ColumnA',u'ColumnB')
        """
        names = []
        for row in self.conn.execute(
                """SELECT name FROM sqlite_master
                WHERE type='table'
                ORDER BY name;
                """):
            names.append(row[0])
        return names

    def get_column_info(self, table_name):
        """return a list of column info

        :param table_name: string; name of the table to query
        :return: list; list of sqlite3.Row

        Example return value:
           [{'cid': 0, 'name': u'Book', 'type': u'INT', 'notnull': 0,
             'dflt_value': None, 'pk': 0}]

        See also: https://www.sqlite.org/pragma.html#pragma_table_info
        """
        info = []
        for row in self.conn.execute(
                "pragma table_info({0})".format(table_name)).fetchall():
            info.append(row)
        return info

    def print_db_info(self):
        """print database layout

        :return: None
        """
        table_info = self.get_db_info()
        for t in table_info:
            print('{}:'.format(t))
            for c in table_info[t].values():
                pk = ' PRIMARY KEY' if c['pk'] else ''
                print('  {}: {}{}'.format(c['name'], c['type'], pk))

    @staticmethod
    def print_column_info(column_info):
        """print the contents of a list returned by get_column_info()

        :param column_info: list
        :return: None
        """
        for name in sorted(column_info.keys()):
            print('{0}:'.format(name))
            info = column_info[name]
            for row in info:
                print('  {0}'.format(row))

    def cursor(self, cursor_class=None):
        """return a cursor; wrapper for sqlite3 method

        :param cursor_class: optional custom cursor_factory class
        :return: sqlite3.Cursor
        """
        if not cursor_class:
            return self.conn.cursor()
        else:
            return self.conn.cursor(cursor_class)

    def execute(self, sql, params=None):
        """execute an arbitrary SQL statement; wrapper for sqlite3 function

        :param sql: string
        :param params: sequence
        :return: sqlite3.Cursor
        """
        if params:
            return self.conn.execute(sql, params)
        else:
            return self.conn.execute(sql)

    def executemany(self, sql, params=None):
        if params:
            return self.conn.executemany(sql, params)
        else:
            return self.conn.executemany(sql)

    def rollback(self, *args, **kwargs):
        """roll back uncommitted transaction; wrapper for sqlite3 function

        :param args: varargs; see documents for sqlite3.commit()
        :param kwargs: kwargs;  see documents for sqlite3.commit()
        :return: sqlite3.Cursor
        """
        return self.conn.rollback(*args, **kwargs)

    def commit(self, *args, **kwargs):
        """commit transaction; wrapper for sqlite3 function

        :param args: varargs; see documents for sqlite3.commit()
        :param kwargs:  kwargs; see documents for sqlite3.commit()
        :return: sqlite3.Cursor
        """
        return self.conn.commit(*args, **kwargs)

    def close(self, *args, **kwargs):
        """roll back uncommitted transaction; wrapper for sqlite3 function

        :param args: varargs; see documents for sqlite3.commit()
        :param kwargs: kwargs;  see documents for sqlite3.commit()
        :return: sqlite3.Cursor
        """
        return self.conn.close(*args, **kwargs)
